import ipywidgets as widgets
import numpy as np
from ase import Atoms
from ase.data import covalent_radii
from ase.data.colors import jmol_colors
from bqplot import LinearScale, ColorScale, Lines, Scatter, Figure, Axis
from bqplot_image_gl import ImageGL
from traitlets import HasTraits, observe, default, List, link, Float, Unicode, Instance, Bool, Int, Any

from abtem.measure import Measurement
from abtem.visualize.interactive.utils import Array


class ColorBar(widgets.HBox):
    min = Float(0.)
    max = Float(1.)
    label = Unicode()

    def __init__(self, color_scale, label='', *args, **kwargs):
        self._x_scale = LinearScale(allow_padding=False)

        scales = {'x': self._x_scale,
                  'y': LinearScale(allow_padding=False, orientation='vertical'),
                  'image': color_scale}

        self._mark = ImageGL(image=np.linspace(0, 1, 500)[None], scales=scales)
        self._x_axis = Axis(scale=scales['x'], label=label)

        # self._x_axis.tick_format = '.2e'
        self._x_axis.num_ticks = 5

        self._figure = Figure(scales=scales,
                              layout=widgets.Layout(height='80px'),
                              axes=[self._x_axis],
                              fig_margin={'top': 0, 'bottom': 50, 'left': 50, 'right': 0})

        self._figure.marks = [self._mark]

        super().__init__(children=[self._figure])

    @property
    def layout(self):
        return self._figure.layout

    @observe('label')
    def _observe_label(self, change):
        self._x_axis.label = change['new']

    @observe('min')
    def _observe_min(self, change):
        self._mark.image = np.linspace(change['new'], self.max, 100)[None]
        self._x_scale.min = change['new']
        self._mark.x = [change['new'], self.max]

    @observe('max')
    def _observe_min(self, change):
        self._mark.image = np.linspace(self.min, change['new'], 100)[None]
        self._x_scale.max = change['new']
        self._mark.x = [self.min, change['new']]


class Artist(HasTraits):
    x_label = Unicode(None, allow_none=True)
    y_label = Unicode(None, allow_none=True)
    visible = Bool()

    def _add_to_canvas(self, canvas):
        raise NotImplementedError()

    @property
    def limits(self):
        raise NotImplementedError()


class ImageArtist(Artist):
    image = Array(check_equal=False)
    extent = List(allow_none=True)
    power = Float(1.)
    color_scheme = Unicode('Greys')
    autoadjust_colorscale = Bool(True)

    def __init__(self, rgb=False, **kwargs):
        self._color_scale = ColorScale(colors=['black', 'white'], min=0, max=1)
        self._color_bar = ColorBar(color_scale=self._color_scale)

        scales = {'x': LinearScale(allow_padding=False),
                  'y': LinearScale(allow_padding=False, orientation='vertical'),
                  'image': self._color_scale}

        if rgb:
            self._mark = ImageGL(image=np.zeros((0, 0, 3)), scales=scales)
        else:
            self._mark = ImageGL(image=np.zeros((0, 0)), scales=scales)

        self._rgb = rgb

        link((self._mark, 'visible'), (self, 'visible'))
        super().__init__(**kwargs)

    @property
    def power_scale_slider(self):
        slider = widgets.FloatSlider(description='Power scale', min=1e-3, max=1, value=1, step=1e-3, )
        slider.observe(self._observe_image, 'value')
        link((slider, 'value'), (self, 'power'))
        return slider

    @property
    def color_scheme_picker(self):
        schemes = ['Greys', 'viridis', 'inferno', 'plasma', 'magma', 'Spectral', 'RdBu']
        dropdown = widgets.Dropdown(description='Scheme', options=schemes)
        link((self, 'color_scheme'), (dropdown, 'value'))
        return dropdown

    @observe('color_scheme')
    def _observe_color_scheme(self, change):
        if change['new'] == 'Greys':
            self._color_scale.colors = ['Black', 'White']
        else:
            self._color_scale.colors = []
            self._color_scale.scheme = change['new']

    @property
    def color_bar(self):
        return self._color_bar

    @default('extent')
    def _default_extent(self):
        return None

    @default('image')
    def _default_image(self):
        return np.zeros((0, 0, 3))

    def _add_to_canvas(self, canvas):
        scales = {'x': canvas.figure.axes[0].scale,
                  'y': canvas.figure.axes[1].scale,
                  'image': self._color_scale}

        self._mark.scales = scales
        canvas.figure.marks = [self._mark] + canvas.figure.marks

    def update_image(self, *args):
        image = self.image
        if self.power != 1:
            image = image ** self.power

        image = np.swapaxes(image, 0, 1)

        if self.extent is None:
            with self._mark.hold_sync():
                self._mark.x = [-.5, image.shape[0] - .5]
                self._mark.y = [-.5, image.shape[1] - .5]
                self._mark.image = image
        else:
            self._mark.image = image

        if self._rgb:
            return

        if (not self.image.size == 0) & self.autoadjust_colorscale:
            with self._mark.hold_sync():
                self._mark.scales['image'].min = float(image.min())
                self._mark.scales['image'].max = float(image.max())

            with self._color_bar._mark.hold_sync():
                self._color_bar.min = float(image.min())
                self._color_bar.max = float(image.max())

    @observe('image')
    def _observe_image(self, *args):
        self.update_image(*args)

    @observe('extent')
    def _observe_extent(self, change):
        if self.image.size == 0:
            return

        sampling = ((change['new'][0][1] - change['new'][0][0]) / self.image.shape[0],
                    (change['new'][1][1] - change['new'][1][0]) / self.image.shape[1])

        self._mark.x = [value - .5 * sampling[0] for value in change['new'][0]]
        self._mark.y = [value - .5 * sampling[1] for value in change['new'][1]]

    @property
    def display_sampling(self):
        if self.image.size == 0:
            return

        extent_x = self._mark.x[1] - self._mark.x[0]
        extent_y = self._mark.y[1] - self._mark.y[0]
        pixel_extent_x = self.image.shape[0]
        pixel_extent_y = self.image.shape[1]
        return (extent_x / pixel_extent_x, extent_y / pixel_extent_y)

    def position_to_index(self, position):
        sampling = self.display_sampling
        px = int(np.round((position[0] - self._mark.x[0]) / sampling[0] - .5))
        py = int(np.round((position[1] - self._mark.y[0]) / sampling[1] - .5))
        return [px, py]

    def indices_to_position(self, indices):
        sampling = self.display_sampling
        px = (indices[0] + .5) * sampling[0] + self._mark.x[0]
        py = (indices[1] + .5) * sampling[1] + self._mark.y[0]
        return [px, py]

    @property
    def limits(self):
        if (self.extent is None) or (self.display_sampling is None):
            return [(-.5, self.image.shape[0] - .5), (-.5, self.image.shape[1] - .5)]

        return [tuple([l - .5 * s for l in L]) for L, s in zip(self.extent, self.display_sampling)]


class ItemSelector(HasTraits):
    sequence = Any()
    current_index = Int(0)
    current_item = Any()

    def __init__(self, **kwargs):
        self._slider = widgets.IntSlider(min=0, max=0, step=1)
        link((self._slider, 'value'), (self, 'current_index'))
        super().__init__(**kwargs)

    @observe('current_index')
    def _observe_current_index(self, change):
        self.current_item = self.sequence[self.current_index]

    @observe('sequence')
    def _observe_sequence(self, change):
        self._slider.max = len(self.sequence) - 1

        clipped_current_index = min(self.current_index, self._slider.max)
        force_trigger = clipped_current_index == self.current_index

        self.current_index = clipped_current_index

        if force_trigger:
            self._observe_current_index(None)


class ArrayViewArtist(Artist):
    array = Array()
    index = Int()

    def __init__(self, **kwargs):
        self._image_artist = ImageArtist()
        super().__init__(**kwargs)

    @property
    def image_artist(self):
        return self._image_artist

    @observe('array', 'index')
    def _observe_array(self, change):
        self._image_artist.image = self.array[self.index]

    def _add_to_canvas(self, canvas):
        self._image_artist._add_to_canvas(canvas)

    @property
    def navigation_sliders(self):
        slider = widgets.IntSlider(min=0, max=len(self.array) - 1, step=1)
        link((slider, 'value'), (self, 'index'))
        return slider

    @property
    def limits(self):
        return self._image_artist.limits


class PointSeriesArtist(Artist):
    points = List()
    index = Int()

    def __init__(self, **kwargs):
        self._scatter_artist = ScatterArtist()
        super().__init__(**kwargs)

    @observe('points', 'index')
    def _observe_points(self, change):
        self._scatter_artist.x = self.points[self.index][:, 0]
        self._scatter_artist.y = self.points[self.index][:, 1]

    def _add_to_canvas(self, canvas):
        self._scatter_artist._add_to_canvas(canvas)

    @property
    def limits(self):
        return self._scatter_artist.limits


class MeasurementArtist2d(Artist):
    measurement = Instance(Measurement, allow_none=True)

    def __init__(self, measurement=None):
        self._image_artist = ImageArtist()
        super().__init__(measurement=measurement)

    @property
    def image_artist(self):
        return self._image_artist

    @observe('measurement')
    def _observe_measurement(self, change):
        if change['new'] is None:
            return

        extent = self.measurement.calibration_limits

        with self._image_artist.hold_trait_notifications():
            self._image_artist.image = change['new'].array  # + np.random.rand()*1e-3
            self._image_artist.extent = extent

        units = self.measurement.calibration_units
        names = self.measurement.calibration_names

        self.x_label = f'{names[0]} [{units[0]}]'
        self.y_label = f'{names[1]} [{units[1]}]'

    def _add_to_canvas(self, canvas):
        self._image_artist._add_to_canvas(canvas)

    @property
    def limits(self):
        return self._image_artist.limits


class Artist1d(Artist):
    x = Array()
    y = Array()

    def __init__(self, mark, color_scale=None, **kwargs):
        self._mark = mark
        self._color_scale = color_scale
        super().__init__(**kwargs)
        link((self._mark, 'visible'), (self, 'visible'))

    @observe('x')
    def _observe_x(self, change):
        self._mark.x = self.x

    @observe('y')
    def _observe_y(self, change):
        self._mark.y = self.y

    def _add_to_canvas(self, canvas):
        scales = {'x': canvas.figure.axes[0].scale,
                  'y': canvas.figure.axes[1].scale}

        try:
            scales['color'] = self._mark.scales['color']
        except KeyError:
            pass

        try:
            scales['size'] = self._mark.scales['size']
        except KeyError:
            pass

        self._mark.scales = scales
        canvas.figure.marks = [self._mark] + canvas.figure.marks

    @property
    def limits(self):
        return [(self.x.min(), self.x.max()), (self.y.min(), self.y.max())]


class ScatterArtist(Artist1d):
    color = Any()

    def __init__(self, colors='red', **kwargs):
        if isinstance(colors, str):
            colors = [colors]

        color_scale = ColorScale(scheme='plasma')

        scales = {'x': LinearScale(allow_padding=False),
                  'y': LinearScale(allow_padding=False, orientation='vertical'),
                  'color': color_scale,
                  'size': LinearScale(min=0, max=1),
                  }
        mark = Scatter(x=np.zeros((1,)), y=np.zeros((1,)), scales=scales, colors=['red'])

        # link((self, 'color'), (mark, 'color'))

        super().__init__(mark=mark, **kwargs)

        link((self._mark, 'visible'), (self, 'visible'))

    @observe('color')
    def _observe_color(self):
        self._mark.color = self.color

    # @default('color')
    # def _default_colors(self):
    #     return ['red']


class LinesArtist(Artist1d):

    def __init__(self, colors='red', **kwargs):
        if isinstance(colors, str):
            colors = [colors]

        scales = {'x': LinearScale(allow_padding=False),
                  'y': LinearScale(allow_padding=False, orientation='vertical'), }

        mark = Lines(x=np.zeros((2,)), y=np.zeros((2,)), scales=scales, colors=colors)

        super().__init__(mark=mark, **kwargs)


class CircleArtist(Artist):
    center = Array()
    radius = Float(1.)

    def __init__(self, **kwargs):
        self._mark = Lines(scales={'x': LinearScale(), 'y': LinearScale()}, colors=['red'])
        super().__init__(**kwargs)
        self.update_mark()

    @default('center')
    def _default_center(self):
        return np.zeros((2,))

    def update_mark(self):
        x = self.center[0] + np.cos(np.linspace(0, 2 * np.pi, 100)) * self.radius
        y = self.center[1] + np.sin(np.linspace(0, 2 * np.pi, 100)) * self.radius
        with self._mark.hold_sync():
            self._mark.x = x
            self._mark.y = y

    @observe('center', 'radius')
    def _observe_center_and_radius(self, change):
        self.update_mark()

    def _add_to_canvas(self, canvas):
        self._mark.scales = {'x': canvas.figure.axes[0].scale, 'y': canvas.figure.axes[1].scale}
        canvas.figure.marks = [self._mark] + canvas.figure.marks

    @property
    def limits(self):
        return [(self.center[0] - self.radius, self.center[0] + self.radius),
                (self.center[1] - self.radius, self.center[1] + self.radius)]


class MeasurementArtist1d(Artist):
    measurement = Instance(Measurement, allow_none=True)

    def __init__(self, measurement=None, **kwargs):
        self._lines_artist = LinesArtist(**kwargs)
        super().__init__(measurement=measurement, **kwargs)

    @property
    def lines_artist(self):
        return self._lines_artist

    @observe('measurement')
    def _observe_measurement(self, change):
        if change['new'] is None:
            return

        with self._lines_artist.hold_trait_notifications():
            self._lines_artist._mark.y = change['new'].array
            self._lines_artist.x = change['new'].calibrations[0].coordinates(len(change['new'].array))

            units = self.measurement.calibration_units
            names = self.measurement.calibration_names

            self.x_label = f'{names[0]} [{units[0]}]'

    def _add_to_canvas(self, canvas):
        self._lines_artist._add_to_canvas(canvas)

    @property
    def limits(self):
        return self._lines_artist.limits


class AtomsArtist(Artist):
    atoms = Instance(Atoms)

    def __init__(self, scale_atoms=500, **kwargs):
        self._scatter_artist = ScatterArtist()
        self._scatter_artist._mark.stroke = 'black'
        self._scale_atoms = scale_atoms
        super().__init__(**kwargs)

        link((self._scatter_artist._mark, 'visible'), (self, 'visible'))

        self.x_label = 'x [Å]'
        self.y_label = 'y [Å]'

    def _add_to_canvas(self, canvas):
        self._scatter_artist._add_to_canvas(canvas)

    @property
    def limits(self):
        return self._scatter_artist.limits

    @observe('atoms')
    def _observe_atoms(self, change):
        x, y = self.atoms.get_positions()[:, :2].T
        self._scatter_artist.x = x
        self._scatter_artist.y = y
        colors = ['#%02x%02x%02x' % tuple((jmol_colors[i] * 255).astype(np.int)) for i in self.atoms.numbers]
        sizes = [int(covalent_radii[i] * self._scale_atoms) for i in self.atoms.numbers]
        self._scatter_artist._mark.colors = colors
        self._scatter_artist._mark.size = sizes



import numpy as np
from scipy.ndimage import center_of_mass

from abtem.device import get_array_module
from abtem.measure import Measurement, calibrations_from_grid
from abtem.utils import ProgressBar, fft_shift
from abtem.waves import Probe


def _run_epie(object,
              probe: np.ndarray,
              diffraction_patterns: np.ndarray,
              positions: np.ndarray,
              maxiter: int,
              alpha: float = 1.,
              beta: float = 1.,
              fix_probe: bool = False,
              fix_com: bool = False,
              return_iterations: bool = False,
              seed=None):
    xp = get_array_module(probe)

    object = xp.array(object)
    probe = xp.array(probe)

    if len(diffraction_patterns.shape) != 3:
        raise ValueError()

    if len(diffraction_patterns) != len(positions):
        raise ValueError()

    if object.shape == (2,):
        object = xp.ones((int(object[0]), int(object[1])), dtype=xp.complex64)
    elif len(object.shape) != 2:
        raise ValueError()

    if probe.shape != diffraction_patterns.shape[1:]:
        raise ValueError()

    if probe.shape != object.shape:
        raise ValueError()

    if return_iterations:
        object_iterations = []
        probe_iterations = []
        SSE_iterations = []

    if seed is not None:
        np.random.seed(seed)

    diffraction_patterns = np.fft.ifftshift(np.sqrt(diffraction_patterns), axes=(-2, -1))

    SSE = 0.
    k = 0
    outer_pbar = ProgressBar(total=maxiter)
    inner_pbar = ProgressBar(total=len(positions))

    while k < maxiter:
        indices = np.arange(len(positions))
        np.random.shuffle(indices)

        old_position = xp.array((0., 0.))
        inner_pbar.reset()
        SSE = 0.
        for j in indices:
            position = xp.array(positions[j])

            diffraction_pattern = xp.array(diffraction_patterns[j])
            illuminated_object = fft_shift(object, old_position - position)

            g = illuminated_object * probe
            gprime = xp.fft.ifft2(diffraction_pattern * xp.exp(1j * xp.angle(xp.fft.fft2(g))))

            object = illuminated_object + alpha * (gprime - g) * xp.conj(probe) / (xp.max(xp.abs(probe)) ** 2)
            old_position = position

            if not fix_probe:
                probe = probe + beta * (gprime - g) * xp.conj(illuminated_object) / (
                        xp.max(xp.abs(illuminated_object)) ** 2)

            # SSE += xp.sum(xp.abs(G) ** 2 - diffraction_pattern) ** 2
            inner_pbar.update(1)

        object = fft_shift(object, position)

        if fix_com:
            com = center_of_mass(xp.fft.fftshift(xp.abs(probe) ** 2))
            probe = xp.fft.ifftshift(fft_shift(probe, - xp.array(com)))

        # SSE = SSE / np.prod(diffraction_patterns.shape)

        if return_iterations:
            object_iterations.append(object)
            probe_iterations.append(probe)
            SSE_iterations.append(SSE)

        outer_pbar.update(1)
        # if verbose:
        #    print(f'Iteration {k:<{len(str(maxiter))}}, SSE = {float(SSE):.3e}')

        k += 1

    inner_pbar.close()
    outer_pbar.close()

    if return_iterations:
        return object_iterations, probe_iterations, SSE_iterations
    else:
        return object, probe, SSE


def epie(measurement: Measurement,
         probe_guess: Probe,
         maxiter: int = 5,
         alpha: float = 1.,
         beta: float = 1.,
         fix_probe: bool = False,
         fix_com: bool = False,
         return_iterations: bool = False,
         max_angle=None,
         seed=None,
         device='cpu', ):
    """
    Reconstruct the phase of a 4D-STEM measurement using the extended Ptychographical Iterative Engine.

    See https://doi.org/10.1016/j.ultramic.2009.05.012

    Parameters
    ----------
    measurement : Measurement object
        4D-STEM measurement.
    probe_guess : Probe object
        The initial guess for the probe.
    maxiter : int
        Run the algorithm for this many iterations.
    alpha : float
        Controls the size of the iterative updates for the object. See reference.
    beta : float
        Controls the size of the iterative updates for the probe. See reference.
    fix_probe : bool
        If True, the probe will not be updated by the algorithm. Default is False.
    fix_com : bool
        If True, the center of mass of the probe will be centered. Default is True.
    return_iterations : bool
        If True, return the reconstruction after every iteration. Default is False.
    max_angle : float, optional
        The maximum reconstructed scattering angle. If this is larger than the input data, the data will be zero-padded.
    seed : int, optional
        Seed the random number generator.
    device : str
        Set the calculation device.

    Returns
    -------
    List of Measurement objects

    """

    diffraction_patterns = measurement.array.reshape((-1,) + measurement.array.shape[2:])

    if max_angle:
        padding_x = int((max_angle / abs(measurement.calibrations[-2].offset) *
                         diffraction_patterns.shape[-2]) // 2) - diffraction_patterns.shape[-2] // 2
        padding_y = int((max_angle / abs(measurement.calibrations[-1].offset) *
                         diffraction_patterns.shape[-1]) // 2) - diffraction_patterns.shape[-1] // 2
        diffraction_patterns = np.pad(diffraction_patterns, ((0,) * 2, (padding_x,) * 2, (padding_y,) * 2))

    extent = (probe_guess.wavelength * 1e3 / measurement.calibrations[2].sampling,
              probe_guess.wavelength * 1e3 / measurement.calibrations[3].sampling)

    sampling = (extent[0] / diffraction_patterns.shape[-2],
                extent[1] / diffraction_patterns.shape[-1])

    x = measurement.calibrations[0].coordinates(measurement.shape[0]) / sampling[0]
    y = measurement.calibrations[1].coordinates(measurement.shape[1]) / sampling[1]
    x, y = np.meshgrid(x, y, indexing='ij')
    positions = np.array([x.ravel(), y.ravel()]).T

    probe_guess.extent = extent
    probe_guess.gpts = diffraction_patterns.shape[-2:]

    calibrations = calibrations_from_grid(probe_guess.gpts, probe_guess.sampling, names=['x', 'y'], units='Å')

    probe_guess._device = device
    probe_guess = probe_guess.build(np.array([0, 0])).array[0]

    result = _run_epie(diffraction_patterns.shape[-2:],
                       probe_guess,
                       diffraction_patterns,
                       positions,
                       maxiter=maxiter,
                       alpha=alpha,
                       beta=beta,
                       return_iterations=return_iterations,
                       fix_probe=fix_probe,
                       fix_com=fix_com,
                       seed=seed)

    if return_iterations:
        object_iterations = [Measurement(object, calibrations=calibrations) for object in result[0]]
        probe_iterations = [Measurement(np.fft.fftshift(probe), calibrations=calibrations) for probe in result[1]]
        return object_iterations, probe_iterations, result[2]
    else:
        return (Measurement(result[0], calibrations=calibrations),
                Measurement(np.fft.fftshift(result[1]), calibrations=calibrations),
                result[2])



"""Module for plotting atoms, images, line scans, and diffraction patterns."""
from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
from ase.data import covalent_radii, chemical_symbols
from ase.data.colors import jmol_colors
from matplotlib.collections import PatchCollection
from matplotlib.patches import Circle
from abtem.visualize.utils import format_label
from typing import Union, Tuple
from matplotlib.lines import Line2D
from abtem.visualize.utils import domain_coloring

#: Array to facilitate the display of cell boundaries.
_cube = np.array([[[0, 0, 0], [0, 0, 1]],
                  [[0, 0, 0], [0, 1, 0]],
                  [[0, 0, 0], [1, 0, 0]],
                  [[0, 0, 1], [0, 1, 1]],
                  [[0, 0, 1], [1, 0, 1]],
                  [[0, 1, 0], [1, 1, 0]],
                  [[0, 1, 0], [0, 1, 1]],
                  [[1, 0, 0], [1, 1, 0]],
                  [[1, 0, 0], [1, 0, 1]],
                  [[0, 1, 1], [1, 1, 1]],
                  [[1, 0, 1], [1, 1, 1]],
                  [[1, 1, 0], [1, 1, 1]]])


def _plane2axes(plane):
    """Internal function for extracting axes from a plane."""
    axes = ()
    last_axis = [0, 1, 2]
    for axis in list(plane):
        if axis == 'x':
            axes += (0,)
            last_axis.remove(0)
        if axis == 'y':
            axes += (1,)
            last_axis.remove(1)
        if axis == 'z':
            axes += (2,)
            last_axis.remove(2)
    return axes + (last_axis[0],)


def show_atoms(atoms, repeat: Tuple[int, int] = (1, 1), scans=None, plane: Union[Tuple[float, float], str] = 'xy',
               ax=None, scale_atoms: float = .5, title: str = None, numbering: bool = False, figsize=None,
               legend=False):
    """
    Show atoms function

    Function to display atoms, especially in Jupyter notebooks.

    Parameters
    ----------
    atoms : ASE atoms object
        The atoms to be shown.
    repeat : two ints, optional
        Tiling of the image. Default is (1,1), ie. no tiling.
    scans : ndarray, optional
        List of scans to apply. Default is None.
    plane : str, two float
        The projection plane given as a combination of 'x' 'y' and 'z', e.g. 'xy', or the as two floats representing the
        azimuth and elevation angles in degrees of the viewing direction, e.g. (45, 45).
    ax : axes object
        pyplot axes object.
    scale_atoms : float
        Scaling factor for the atom display sizes. Default is 0.5.
    title : str
        Title of the displayed image. Default is None.
    numbering : bool
        Option to set plot numbering. Default is False.
    """

    atoms = atoms.copy()
    atoms *= repeat + (1,)

    if isinstance(plane, str):
        ax = _show_atoms_2d(atoms, scans, plane, ax, scale_atoms, title, numbering, figsize, legend=legend)
    else:
        if scans is not None:
            raise NotImplementedError()

        if numbering:
            raise NotImplementedError()
        ax = _show_atoms_3d(atoms, plane[0], plane[1], scale_atoms=scale_atoms, ax=ax, figsize=figsize)

    return ax


def _show_atoms_2d(atoms, scans=None, plane: Union[Tuple[float, float], str] = 'xy', ax=None, scale_atoms: float = .5,
                   title: str = None, numbering: bool = False, figsize=None, legend=False):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    cell = atoms.cell
    axes = _plane2axes(plane)

    for line in _cube:
        cell_lines = np.array([np.dot(line[0], cell), np.dot(line[1], cell)])
        ax.plot(cell_lines[:, axes[0]], cell_lines[:, axes[1]], 'k-')

    if len(atoms) > 0:
        positions = atoms.positions[:, axes[:2]]
        order = np.argsort(atoms.positions[:, axes[2]])
        positions = positions[order]

        colors = jmol_colors[atoms.numbers[order]]
        sizes = covalent_radii[atoms.numbers[order]] * scale_atoms

        circles = []
        for position, size in zip(positions, sizes):
            circles.append(Circle(position, size))

        coll = PatchCollection(circles, facecolors=colors, edgecolors='black')
        ax.add_collection(coll)

        ax.axis('equal')
        ax.set_xlabel(plane[0] + ' [Å]')
        ax.set_ylabel(plane[1] + ' [Å]')

        ax.set_title(title)

        if numbering:
            for i, (position, size) in enumerate(zip(positions, sizes)):
                ax.annotate('{}'.format(order[i]), xy=position, ha="center", va="center")

    if legend:
        legend_elements = [Line2D([0], [0], marker='o', color='w', markeredgecolor='k', label=chemical_symbols[unique],
                                  markerfacecolor=jmol_colors[unique], markersize=12)
                           for unique in np.unique(atoms.numbers)]

        ax.legend(handles=legend_elements)

    if scans is not None:
        if not isinstance(scans, Iterable):
            scans = [scans]

        for scan in scans:
            scan.add_to_mpl_plot(ax)

    return ax


def _show_atoms_3d(atoms, azimuth=45., elevation=30., ax=None, scale_atoms=500., margin=1., figsize=None):
    cell = atoms.cell
    colors = jmol_colors[atoms.numbers]
    sizes = covalent_radii[atoms.numbers] ** 2 * scale_atoms
    positions = atoms.positions

    for line in _cube:
        cell_lines = np.array([np.dot(line[0], cell), np.dot(line[1], cell)])
        start = cell_lines[0]
        end = cell_lines[1]
        cell_line_points = start + (end - start)[None] * np.linspace(0, 1, 100)[:, None]
        positions = np.vstack((positions, cell_line_points))
        sizes = np.concatenate((sizes, [1] * len(cell_line_points)))
        colors = np.vstack((colors, [(0, 0, 0)] * len(cell_line_points)))

    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(projection='3d', proj_type='ortho')

    ax.scatter(positions[:, 0],
               positions[:, 1],
               positions[:, 2],
               c=colors,
               marker='o',
               s=sizes,
               alpha=1,
               linewidth=1,
               edgecolor='k')

    xmin = min(min(atoms.positions[:, 0]), min(atoms.cell[:, 0])) - margin
    xmax = max(max(atoms.positions[:, 0]), max(atoms.cell[:, 0])) + margin
    ymin = min(min(atoms.positions[:, 1]), min(atoms.cell[:, 1])) - margin
    ymax = max(max(atoms.positions[:, 1]), max(atoms.cell[:, 1])) + margin
    zmin = min(min(atoms.positions[:, 2]), min(atoms.cell[:, 2])) - margin
    zmax = max(max(atoms.positions[:, 2]), max(atoms.cell[:, 2])) + margin

    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])
    ax.set_zlim([zmin, zmax])

    ax.set_xlabel('x [Å]')
    ax.set_ylabel('y [Å]')
    ax.set_zlabel('z [Å]')

    ax.grid(False)

    ax.azim = azimuth
    ax.elev = elevation

    ax.set_box_aspect([xmax - xmin, ymax - ymin, zmax - zmin])
    return ax


def show_measurement_2d(measurement,
                        ax=None,
                        figsize=None,
                        cbar=False,
                        cbar_label=None,
                        cmap='gray',
                        discrete_cmap=False,
                        vmin=None,
                        vmax=None,
                        power=1.,
                        log_scale=False,
                        title=None,
                        equal_ticks=False,
                        is_rgb=False,
                        x_label=None,
                        y_label=None,
                        **kwargs):
    """
    Show image function

    Function to display an image.

    Parameters
    ----------
    array : ndarray
        Image array.
    calibrations : tuple of calibration objects.
        Spatial calibrations.
    ax : axes object
        pyplot axes object.
    title : str, optional
        Image title. Default is None.
    colorbar : bool, optional
        Option to show a colorbar. Default is False.
    cmap : str, optional
        Colormap name. Default is 'gray'.
    figsize : float, pair of float, optional
        Size of the figure in inches, either as a square for one number or a rectangle for two. Default is None.
    scans : ndarray, optional
        Array of scans. Default is None.
    discrete : bool, optional
        Option to discretize intensity values to integers. Default is False.
    cbar_label : str, optional
        Text label for the color bar. Default is None.
    vmin : float, optional
        Minimum of the intensity scale. Default is None.
    vmax : float, optional
        Maximum of the intensity scale. Default is None.
    kwargs :
        Remaining keyword arguments are passed to pyplot.
    """

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if is_rgb:
        calibrations = measurement.calibrations[-3:-1]
    else:
        calibrations = measurement.calibrations[-2:]

    if not is_rgb:
        array = measurement.array[(0,) * (measurement.dimensions - 2) + (slice(None),) * 2]
    else:
        array = measurement.array[:, :, :]

    if np.iscomplexobj(array):
        array = domain_coloring(array)

    if power != 1:
        array = array ** power

    if log_scale:
        array = np.log(array)

    extent = []
    for calibration, num_elem in zip(calibrations, array.shape):
        extent.append(calibration.offset)
        extent.append(calibration.offset + num_elem * calibration.sampling - calibration.sampling)

    if vmin is None:
        vmin = np.min(array)
        if discrete_cmap:
            vmin -= .5

    if vmax is None:
        vmax = np.max(array)
        if discrete_cmap:
            vmax += .5

    if discrete_cmap:
        cmap = plt.get_cmap(cmap, np.max(array) - np.min(array) + 1)

    im = ax.imshow(np.swapaxes(array, 0, 1), extent=extent, cmap=cmap, origin='lower', vmin=vmin, vmax=vmax,
                   interpolation='nearest',
                   **kwargs)

    if cbar:
        if cbar_label is None:
            cbar_label = format_label(measurement)

        cax = plt.colorbar(im, ax=ax, label=cbar_label)
        if discrete_cmap:
            cax.set_ticks(ticks=np.arange(np.min(array), np.max(array) + 1))

    if x_label is None:
        x_label = format_label(calibrations[-2])

    if y_label is None:
        y_label = format_label(calibrations[-1])

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if title is not None:
        ax.set_title(title)
    elif len(measurement.array.shape) > 2:
        if any([n > 1 for n in measurement.array.shape[:-2]]):
            ax.set_title(f'Slice {(0,) * (len(measurement.array.shape) - 2)} of {measurement.array.shape} measurement')

    if equal_ticks:
        d = max(np.diff(ax.get_xticks())[0], np.diff(ax.get_yticks())[0])
        xticks = np.arange(*ax.get_xlim(), d)
        yticks = np.arange(*ax.get_ylim(), d)
        ax.set_xticks(xticks)
        ax.set_yticks(yticks)

    return ax, im


def show_measurement_1d(measurement, ax=None, figsize=None, legend=False, title=None, label=None,
                        x_label=None, y_label=None, **kwargs):
    """
    Show line function

    Function to display a line scan.

    Parameters
    ----------
    array : ndarray
        Array of measurement values along a line.
    calibration : calibration object
        Spatial calibration for the line.
    ax : axes object, optional
        pyplot axes object.
    title : str, optional
        Title for the plot. Default is None.
    legend : bool, optional
        Option to display a plot legend. Default is False.
    kwargs :
       Remaining keyword arguments are passed to pyplot.
    """

    calibration = measurement.calibrations[0]
    array = measurement.array
    if calibration is None:
        x = np.arange(len(array))
    else:
        x = np.linspace(calibration.offset, calibration.offset + len(array) * calibration.sampling, len(array))

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if not label:
        label = measurement.name

    lines = ax.plot(x, array, label=label, **kwargs)

    if x_label is None:
        x_label = format_label(calibration)

    if y_label is None:
        y_label = format_label(measurement)

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if legend:
        ax.legend()

    if title is not None:
        ax.set_title(title)

    return ax, lines[0]



import os

import pytest
from ase.io import read

_ROOT = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture(scope='session')
def data_path():
    """Internal function to set the parametrization data directory."""
    return os.path.join(_ROOT, 'data')

@pytest.fixture(scope='session')
def graphene_atoms(data_path):
    return read(os.path.join(data_path, 'orthogonal_graphene.cif'))


def pytest_configure(config):
    config.addinivalue_line("markers", "gpaw")
    config.addinivalue_line("markers", "gpu")


def pytest_addoption(parser):
    parser.addoption("--rungpaw", action="store_true", help="run gpaw tests")
    parser.addoption("--rungpu", action="store_true", help="run gpu tests")


def pytest_runtest_setup(item):
    if 'gpaw' in item.keywords and not item.config.getvalue("rungpaw"):
        pytest.skip("need --rungpaw option to run")

    if 'gpu' in item.keywords and not item.config.getvalue("rungpu"):
        pytest.skip("need --rungpu option to run")



import os
from io import open as io_open

from setuptools import setup, find_packages

__version__ = None
src_dir = os.path.abspath(os.path.dirname(__file__))
version_file = os.path.join(src_dir, 'abtem', '_version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())

fndoc = os.path.join(src_dir, 'README.md')
with io_open(fndoc, mode='r', encoding='utf-8') as fd:
    README_md = fd.read()

setup(
    name='abtem',
    version=__version__,
    description='ab initio Transmission Electron Microscopy',
    long_description=README_md,
    url='https://github.com/jacobjma/abtem',
    author='Jacob Madsen',
    author_email='jacob.madsen@univie.ac.at',
    maintainer='Jacob Madsen',
    maintainer_email='jacob.madsen@univie.ac.at',
    platforms=['any'],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'scipy',
        'numba',
        'imageio',
        'pyfftw',
        'h5py',
        'matplotlib',
        'ase',
        'tqdm',
        'psutil'],
    extras_require={
        'full': ['hyperspy',
                 'pyxem']
    },
    tests_require=['pytest'],
    packages=find_packages(),
    include_package_data=True,
)



"""Module for modifying ASE atoms objects for use in abTEM."""
from fractions import Fraction
from numbers import Number
from typing import Union, Sequence

import numpy as np
from ase import Atoms
from scipy.linalg import polar


def is_cell_hexagonal(atoms: Atoms):
    """
    Function to check whether the cell of an ASE atoms object is hexagonal.

    Parameters
    ----------
    atoms : ASE atoms object
        The atoms that should be checked.
    """
    cell = atoms.get_cell()

    a = np.linalg.norm(cell[0], axis=0)
    b = np.linalg.norm(cell[1], axis=0)
    c = np.linalg.norm(cell[2], axis=0)
    angle = np.arccos(np.dot(cell[0], cell[1]) / (a * b))

    return np.isclose(a, b) & (np.isclose(angle, np.pi / 3) | np.isclose(angle, 2 * np.pi / 3)) & (c == cell[2, 2])


def is_cell_orthogonal(atoms: Atoms, tol: float = 1e-12):
    """
    Check whether an Atoms object has an orthogonal cell.

    Parameters
    ----------
    atoms : ASE atoms object
        The atoms that should be checked.
    tol : float
        Components of the lattice vectors below this value are considered to be zero.
    """
    return not np.any(np.abs(atoms.cell[~np.eye(3, dtype=bool)]) > tol)


def is_cell_valid(atoms: Atoms, tol: float = 1e-12) -> bool:
    """
    Check whether the cell of an ASE atoms object can be converted to a structure that is usable by abTEM.

    Parameters
    ----------
    atoms : ASE atoms object
        The atoms that should be checked.
    tol : float
        Components of the lattice vectors below this value are considered to be zero.

    Returns
    -------
    bool
        If true, the atomic structure is usable by abTEM.
    """
    if np.abs(atoms.cell[0, 0] - np.linalg.norm(atoms.cell[0])) > tol:
        return False

    if np.abs(atoms.cell[1, 2]) > tol:
        return False

    if np.abs(atoms.cell[2, 2] - np.linalg.norm(atoms.cell[2])) > tol:
        return False

    return True


def standardize_cell(atoms: Atoms, tol: float = 1e-12):
    """
    Standardize the cell of an ASE atoms object. The atoms are rotated so one of the lattice vectors in the xy-plane
    aligns with the x-axis, then all of the lattice vectors are made positive.

    Parameters
    ----------
    atoms : ASE atoms object
        The atoms that should be standardized
    tol : float
        Components of the lattice vectors below this value are considered to be zero.

    Returns
    -------
    atoms : ASE atoms object
        The standardized atoms.
    """
    cell = np.array(atoms.cell)

    vertical_vector = np.where(np.all(np.abs(cell[:, :2]) < tol, axis=1))[0]

    if len(vertical_vector) != 1:
        raise RuntimeError('Invalid cell: no vertical lattice vector')

    cell[[vertical_vector[0], 2]] = cell[[2, vertical_vector[0]]]
    r = np.arctan2(atoms.cell[0, 1], atoms.cell[0, 0]) / np.pi * 180

    atoms.set_cell(cell)

    if r != 0.:
        atoms.rotate(-r, 'z', rotate_cell=True)

    if not is_cell_valid(atoms, tol):
        raise RuntimeError('This cell cannot be made orthogonal using currently implemented methods.')

    atoms.set_cell(np.abs(atoms.get_cell()))

    atoms.wrap()
    return atoms


def orthogonalize_cell(atoms: Atoms, limit_denominator: int = 10, preserve_periodicity: bool = True,
                       return_strain: bool = False):
    """
    Make the cell of an ASE atoms object orthogonal. This is accomplished by repeating the cell until the x-component
    of the lattice vectors in the xy-plane closely matches. If the ratio between the x-components is irrational this
    may not be possible without introducing some strain. However, the amount of strain can be made arbitrarily small
    by using many repetitions.

    Parameters
    ----------
    atoms : ASE atoms object
        The non-orthogonal atoms object.
    limit_denominator : int
        The maximum denominator in the rational approximation. Increase this to allow more repetitions and hence less
        strain.
    preserve_periodicity : bool, optional
        This function will make a structure periodic while preserving periodicity exactly, this will generally result in
        repeating the structure. If preserving periodicity is not desired, this may be set to False. Default is True.
    return_strain : bool
        If true, return the strain tensor that were applied to make the atoms orthogonal.

    Returns
    -------
    atoms : ASE atoms object
        The orthogonal atoms.
    strain_tensor : 2x2 array
        The applied strain tensor. Only provided if return_strain is true.
    """
    if is_cell_orthogonal(atoms):
        return atoms

    atoms = atoms.copy()
    atoms = standardize_cell(atoms)

    if not preserve_periodicity:
        return cut_rectangle(atoms, origin=(0, 0), extent=np.diag(atoms.cell)[:2])

    fraction = atoms.cell[0, 0] / atoms.cell[1, 0]
    fraction = Fraction(fraction).limit_denominator(limit_denominator)

    atoms *= (fraction.denominator, fraction.numerator, 1)

    new_cell = atoms.cell.copy()
    new_cell[1, 0] = new_cell[0, 0]

    a = np.linalg.solve(atoms.cell[:2, :2], new_cell[:2, :2])
    _, strain_tensor = polar(a, side='left')
    strain_tensor[0, 0] -= 1
    strain_tensor[1, 1] -= 1

    atoms.set_cell(new_cell, scale_atoms=True)
    atoms.set_cell(np.diag(atoms.cell))
    atoms.wrap()

    if return_strain:
        return atoms, strain_tensor
    else:
        return atoms


def cut_rectangle(atoms: Atoms, origin: Sequence[float], extent: Sequence[float], margin: float = 0.):
    """
    Cuts out a cell starting at the origin to a given extent from a sufficiently repeated copy of atoms.

    Parameters
    ----------
    atoms : ASE atoms object
        This should correspond to a repeatable unit cell.
    origin : two float
        Origin of the new cell. Units of Angstrom.
    extent : two float
        xy-extent of the new cell. Units of Angstrom.
    margin : float
        Atoms within margin from the border of the new cell will be included. Units of Angstrom. Default is 0.

    Returns
    -------
    ASE atoms object
    """

    # TODO : check that this works in edge cases

    atoms = atoms.copy()
    cell = atoms.cell.copy()

    extent = (extent[0], extent[1], atoms.cell[2, 2],)
    atoms.positions[:, :2] -= np.array(origin)

    a = atoms.cell.scaled_positions(np.array((extent[0] + 2 * margin, 0, 0)))
    b = atoms.cell.scaled_positions(np.array((0, extent[1] + 2 * margin, 0)))

    repetitions = (int(np.ceil(abs(a[0])) + np.ceil(abs(b[0]))),
                   int(np.ceil(abs(a[1])) + np.ceil(abs(b[1]))), 1)

    shift = (-np.floor(min(a[0], 0)) - np.floor(min(b[0], 0)),
             -np.floor(min(a[1], 0)) - np.floor(min(b[1], 0)), 0)
    atoms.set_scaled_positions(atoms.get_scaled_positions() - shift)

    atoms *= repetitions

    atoms.positions[:, :2] -= margin

    atoms.set_cell([extent[0], extent[1], cell[2, 2]])

    atoms = atoms[((atoms.positions[:, 0] >= -margin) &
                   (atoms.positions[:, 1] >= -margin) &
                   (atoms.positions[:, 0] < extent[0] + margin) &
                   (atoms.positions[:, 1] < extent[1] + margin))
    ]
    return atoms


def pad_atoms(atoms: Atoms, margin: float, directions='xy', in_place=False):
    """
    Repeat the atoms in x and y, retaining only the repeated atoms within the margin distance from the cell boundary.

    Parameters
    ----------
    atoms: ASE Atoms object
        The atoms that should be padded.
    margin: float
        The padding margin.

    Returns
    -------
    ASE Atoms object
        Padded atoms.
    """

    if not is_cell_orthogonal(atoms):
        raise RuntimeError('The cell of the atoms must be orthogonal.')

    if not in_place:
        atoms = atoms.copy()

    old_cell = atoms.cell.copy()

    axes = [{'x': 0, 'y': 1, 'z': 2}[direction] for direction in directions]

    reps = [1, 1, 1]
    for axis in axes:
        reps[axis] = int(1 + 2 * np.ceil(margin / atoms.cell[axis, axis]))

    if any([rep > 1 for rep in reps]):
        atoms *= reps
        atoms.positions[:] -= np.diag(old_cell) * [rep // 2 for rep in reps]
        atoms.cell = old_cell

    # import matplotlib.pyplot as plt
    # from abtem import show_atoms
    # show_atoms(atoms, plane='xz')
    # plt.show()

    to_keep = np.ones(len(atoms), dtype=bool)
    for axis in axes:
        to_keep *= (atoms.positions[:, axis] > -margin) * (atoms.positions[:, axis] < atoms.cell[axis, axis] + margin)

    atoms = atoms[to_keep]

    # for axis in axes:
    #     left = atoms[atoms.positions[:, axis] < margin]
    #     left.positions[:, axis] += atoms.cell[axis, axis]
    #     right = atoms[(atoms.positions[:, axis] > atoms.cell[axis, axis] - margin) &
    #                   (atoms.positions[:, axis] < atoms.cell[axis, axis])]
    #     right.positions[:, axis] -= atoms.cell[axis, axis]
    #     atoms += left + right
    return atoms


def flip_atoms(atoms):
    atoms = atoms.copy()
    atoms.positions[:] = atoms.cell[2, 2] - atoms.positions[:]
    return atoms


class SlicedAtoms:

    def __init__(self, atoms, slice_thicknesses):
        self._atoms = atoms
        self.slice_thicknesses = slice_thicknesses

    @property
    def atoms(self):
        return self._atoms

    @property
    def positions(self):
        return self.atoms.positions

    @property
    def numbers(self):
        return self.atoms.numbers

    def __len__(self):
        return len(self.atoms)

    @property
    def slice_thicknesses(self):
        return self._slice_thicknesses

    @slice_thicknesses.setter
    def slice_thicknesses(self, slice_thicknesses):
        if isinstance(slice_thicknesses, Number):
            num_slices = int(np.ceil(self._atoms.cell[2, 2] / slice_thicknesses))
            slice_thicknesses = np.full(num_slices, float(slice_thicknesses))
        self._slice_thicknesses = slice_thicknesses

    def flip(self):
        self._atoms.positions[:] = self._atoms.cell[2, 2] - self._atoms.positions[:]
        self._slice_thicknesses[:] = self._slice_thicknesses[::-1]

    def get_slice_entrance(self, i):
        return max(np.sum(self.slice_thicknesses[:i]), 0)

    def get_slice_exit(self, i):
        return min(self.get_slice_entrance(i) + self.slice_thicknesses[i], self.atoms.cell[2, 2])

    def get_subsliced_atoms(self,
                            start,
                            end=None,
                            atomic_number=None,
                            padding: Union[bool, float] = False,
                            z_margin=0.):

        if end is None:
            end = start + 1

        a = self.get_slice_entrance(start) - z_margin
        b = self.get_slice_entrance(end) + z_margin

        in_slice = (self.atoms.positions[:, 2] >= a) * (self.atoms.positions[:, 2] < b)

        if atomic_number is not None:
            in_slice = (self.atoms.numbers == atomic_number) * in_slice

        atoms = self.atoms[in_slice]

        if padding:
            atoms = pad_atoms(atoms, padding)

        return self.__class__(atoms, self.slice_thicknesses)

    @property
    def num_slices(self):
        """The number of projected potential slices."""
        return len(self._slice_thicknesses)

    def get_slice_thickness(self, i):
        """The thickness of the projected potential slices."""
        return self._slice_thicknesses[i]



import os

import numpy as np
import pytest
from abtem.detect import AnnularDetector, PixelatedDetector
from abtem.potentials import Potential
from abtem.scan import PositionScan, LineScan, GridScan
from abtem.waves import Probe
from ase.io import read


def test_custom_scan():
    positions = np.array([(2, 2), (1, 1)])
    scan = PositionScan(positions=positions)

    assert np.all(next(scan.generate_positions(1))[1] == positions[0])
    assert np.all(next(scan.generate_positions(2))[1] == positions)


def test_line_scan():
    start = (0, 0)
    end = (1, 1)
    scan = LineScan(start, end, gpts=5)

    positions = scan.get_positions()
    assert np.allclose(positions[0], start)
    assert np.allclose(positions[-1], end)
    assert np.allclose(positions[2], np.mean([start, end], axis=0))
    assert np.allclose(np.linalg.norm(np.diff(positions, axis=0), axis=1), scan.sampling[0])

    scan = LineScan(start, end, gpts=5, endpoint=False)
    positions = scan.get_positions()
    assert np.allclose(positions[0], start)
    assert np.allclose(positions[-1], (end[0] - (end[0] - start[0]) / 5, end[1] - (end[1] - start[1]) / 5))


def test_gridscan_raises():
    with pytest.raises(ValueError) as e:
        GridScan(start=0, end=1)

    assert str(e.value) == 'Scan start/end has incorrect shape'


def test_grid_scan():
    start = (0, 0)
    end = (1, 2)
    scan = GridScan(start, end, gpts=5, endpoint=True)

    positions = scan.get_positions()
    assert np.all(positions[0] == start)
    assert np.all(positions[-1] == end)
    assert np.allclose(positions[4], [start[1], end[1]])
    assert np.allclose(np.linalg.norm(np.diff(positions[:4], axis=0), axis=1), scan.sampling[1])

    scan = GridScan(start, end, gpts=5, endpoint=False)
    positions = scan.get_positions()
    assert np.all(positions[0] == start)
    assert np.allclose(positions[-1], (end[0] - (end[0] - start[0]) / 5, end[1] - (end[1] - start[1]) / 5))


def test_partition():
    gridscan = GridScan(start=(0, 0), end=(2, 2), sampling=.5, endpoint=False)
    scans = gridscan.partition_scan((2, 2))
    positions = []
    for scan in scans:
        positions.append(scan.get_positions())

    assert np.allclose(((np.vstack(positions)[None] - gridscan.get_positions()[:, None]) ** 2).min(0), 0)
    assert np.allclose(((np.vstack(positions)[None] - gridscan.get_positions()[:, None]) ** 2).min(1), 0)

    gridscan = GridScan(start=(0, 0), end=(2, 2), sampling=.5, endpoint=True)
    scans = gridscan.partition_scan((2, 2))
    positions = []
    for scan in scans:
        positions.append(scan.get_positions())

    assert np.allclose(((np.vstack(positions)[None] - gridscan.get_positions()[:, None]) ** 2).min(0), 0)
    assert np.allclose(((np.vstack(positions)[None] - gridscan.get_positions()[:, None]) ** 2).min(1), 0)


def test_partition_measurement():
    atoms = read(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/amorphous_carbon.cif'))
    potential = Potential(atoms, gpts=256, slice_thickness=1, projection='infinite',
                          parametrization='kirkland').build(pbar=False)

    detector = AnnularDetector(inner=70, outer=100)
    gridscan = GridScan(start=[0, 0], end=potential.extent, gpts=4)

    probe = Probe(semiangle_cutoff=15, energy=300e3)

    measurements = probe.scan(gridscan, detector, potential, pbar=False)

    scans = gridscan.partition_scan((2, 2))
    partitioned_measurements = detector.allocate_measurement(probe, gridscan)

    for scan in scans:
        probe.scan(scan, detector, potential, measurements=partitioned_measurements, pbar=False)

    assert np.allclose(partitioned_measurements.array, measurements.array)


def test_preallocated_measurement():
    atoms = read(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/amorphous_carbon.cif'))
    potential = Potential(atoms, gpts=256, slice_thickness=1, projection='infinite',
                          parametrization='kirkland').build(pbar=False)
    scan = GridScan(start=[0, 0], end=potential.extent, gpts=4)

    detector1 = AnnularDetector(inner=70, outer=100)
    probe = Probe(semiangle_cutoff=15, energy=300e3, extent=potential.extent, gpts=512)

    measurement = detector1.allocate_measurement(probe, scan)
    probe.scan(scan, detector1, potential, measurement, pbar=False)

    assert np.any(measurement.array > 0)

    detector2 = PixelatedDetector()

    measurement1 = detector1.allocate_measurement(probe, scan)
    measurement2 = detector2.allocate_measurement(probe, scan)

    with pytest.raises(ValueError) as e:
        probe.scan(scan, [detector1, detector2], potential, measurement1, pbar=False)

    probe.scan(scan, [detector1, detector2], potential, {detector1: measurement1, detector2: measurement2}, pbar=False)
